@@ -12,6 +12,7 @@
 #include <arpa/inet.h>
 
 #include <osmocom/core/fsm.h>
+#include <osmocom/core/linuxlist.h>
 #include <osmocom/core/utils.h>
 #include <osmocom/core/timer.h>
 #include <osmocom/core/prim.h>
@@ -22,9 +23,12 @@
 #include "xua_msg.h"
 #include <osmocom/sigtran/protocol/sua.h>
 #include <osmocom/sigtran/protocol/m3ua.h>
+#include <sys/types.h>
 
+#include "sccp_internal.h"
 #include "ss7_as.h"
 #include "ss7_asp.h"
+#include "ss7_combined_linkset.h"
 #include "ss7_route.h"
 #include "ss7_route_table.h"
 #include "xua_asp_fsm.h"
@@ -39,7 +43,15 @@ static struct msgb *encode_notify(const struct osmo_xlm_prim_notify *npar)
 	return msg;
 }
 
-static void tx_notify(struct osmo_ss7_asp *asp, const struct osmo_xlm_prim_notify *npar)
+static int fill_notify_route_ctx(const struct osmo_ss7_asp *asp, struct osmo_xlm_prim_notify *npar)
+{
+	npar->route_ctx_count = ss7_asp_get_all_rctx(asp, npar->route_ctx, ARRAY_SIZE(npar->route_ctx), NULL);
+	if (npar->route_ctx_count > 0)
+		npar->presence |= NOTIFY_PAR_P_ROUTE_CTX;
+	return 0;
+}
+
+static void tx_notify(struct osmo_ss7_asp *asp, struct osmo_xlm_prim_notify *npar)
 {
 	const char *type_name, *info_name, *info_str;
 	type_name = get_value_string(m3ua_ntfy_type_names, npar->status_type);
@@ -48,10 +60,46 @@ static void tx_notify(struct osmo_ss7_asp *asp, const struct osmo_xlm_prim_notif
 
 	LOGPASP(asp, DLSS7, LOGL_INFO, "Tx NOTIFY Type %s:%s (%s)\n",
 		type_name, info_name, info_str);
+	fill_notify_route_ctx(asp, npar);
 	struct msgb *msg = encode_notify(npar);
 	osmo_ss7_asp_send(asp, msg);
 }
 
+/* RFC 4666 4.5.1: "For the particular case that an ASP becomes active for an AS and
+ * destinations normally accessible to the AS are inaccessible, restricted, or congested,
+ * the SG MAY send DUNA, DRST, or SCON messages for the inaccessible, restricted, or
+ * congested destinations to the ASP newly active for the AS to prevent the ASP from
+ * sending traffic for destinations that it might not otherwise know that are inaccessible,
+ * restricted, or congested" */
+static void as_tx_duna_during_asp_act(struct osmo_ss7_as *as, struct osmo_ss7_asp *asp)
+{
+	struct osmo_ss7_instance *inst = as->inst;
+	struct osmo_ss7_route_table *rtbl = inst->rtable_system;
+	struct osmo_ss7_as *as_it;
+	uint32_t aff_pc[32];
+	unsigned int num_aff_pc = 0;
+	uint32_t rctx_be = htonl(as->cfg.routing_key.context);
+
+	/* Send up to 32 PC per DUNA: */
+	llist_for_each_entry(as_it, &inst->as_list, list) {
+		if (as == as_it)
+			continue;
+		if (ss7_route_table_dpc_is_accessible_skip_as(rtbl, as_it->cfg.routing_key.pc, as))
+			continue;
+		aff_pc[num_aff_pc++] = htonl(as_it->cfg.routing_key.pc); /* mask = 0 */
+		if (num_aff_pc == ARRAY_SIZE(aff_pc)) {
+			xua_tx_snm_available(asp, &rctx_be, 1,
+					     aff_pc, num_aff_pc,
+					     "RFC4666 4.5.1", false);
+			num_aff_pc = 0;
+		}
+	}
+	if (num_aff_pc > 0)
+		xua_tx_snm_available(asp, &rctx_be, 1,
+				     aff_pc, num_aff_pc,
+				     "RFC4666 4.5.1", false);
+}
+
 static int as_notify_all_asp(struct osmo_ss7_as *as, struct osmo_xlm_prim_notify *npar)
 {
 	struct msgb *msg;
@@ -83,9 +131,9 @@ static int as_notify_all_asp(struct osmo_ss7_as *as, struct osmo_xlm_prim_notify
 			continue;
 
 		/* Optional: ASP Identifier (if sent in ASP-UP) */
-		if (asp->asp_id_present) {
+		if (asp->remote_asp_id_present) {
 			npar->presence |= NOTIFY_PAR_P_ASP_ID;
-			npar->asp_id = asp->asp_id;
+			npar->asp_id = asp->remote_asp_id;
 		} else
 			npar->presence &= ~NOTIFY_PAR_P_ASP_ID;
 
@@ -93,6 +141,7 @@ static int as_notify_all_asp(struct osmo_ss7_as *as, struct osmo_xlm_prim_notify
 
 		LOGPASP(asp, DLSS7, LOGL_INFO, "Tx NOTIFY Type %s:%s (%s)\n",
 			type_name, info_name, info_str);
+		fill_notify_route_ctx(asp, npar);
 		msg = encode_notify(npar);
 		osmo_ss7_asp_send(asp, msg);
 		sent++;
@@ -101,26 +150,6 @@ static int as_notify_all_asp(struct osmo_ss7_as *as, struct osmo_xlm_prim_notify
 	return sent;
 }
 
-/* determine which role (SG/ASP/IPSP) we operate in */
-static int get_local_role(struct osmo_ss7_as *as)
-{
-	unsigned int i;
-
-	/* this is a bit tricky. "osmo_ss7_as" has no configuration of a role,
-	 * only the ASPs have.  As they all must be of the same role, let's simply
-	 * find the first one and return its role */
-	for (i = 0; i < ARRAY_SIZE(as->cfg.asps); i++) {
-		struct osmo_ss7_asp *asp = as->cfg.asps[i];
-
-		if (!asp)
-			continue;
-
-		return asp->cfg.role;
-	}
-	/* we don't have any ASPs in this AS? Strange */
-	return -1;
-}
-
 static struct msgb *xua_as_encode_msg(const struct osmo_ss7_as *as, struct xua_msg *xua)
 {
 	switch (as->cfg.proto) {
@@ -172,7 +201,7 @@ int xua_as_transmit_msg(struct osmo_ss7_as *as, struct xua_msg *xua)
 		 * strictly required by M3UA RFC, but would fit the overall
 		 * principle. */
 	case OSMO_SS7_AS_TMOD_ROUNDROBIN:
-		asp = ss7_as_select_asp(as, &xua->mtp);
+		asp = ss7_as_select_asp(as, xua);
 		break;
 	case OSMO_SS7_AS_TMOD_BCAST:
 		return xua_as_transmit_msg_broadcast(as, xua);
@@ -212,17 +241,14 @@ static const struct value_string xua_as_event_names[] = {
 
 struct xua_as_fsm_priv {
 	struct osmo_ss7_as *as;
-	struct {
+	struct { /* RFC4666 recovery timer T(r) */
 		struct osmo_timer_list t_r;
 		struct llist_head queued_xua_msgs;
 	} recovery;
-	bool ipa_route_created;
 };
 
 static void fill_notify_statchg_pars(const struct osmo_fsm_inst *fi, struct osmo_xlm_prim_notify *npar)
 {
-	struct xua_as_fsm_priv *xafp = (struct xua_as_fsm_priv *) fi->priv;
-	struct osmo_ss7_as *as = xafp->as;
 	*npar = (struct osmo_xlm_prim_notify){
 		.status_type = M3UA_NOTIFY_T_STATCHG,
 	};
@@ -242,83 +268,8 @@ static void fill_notify_statchg_pars(const struct osmo_fsm_inst *fi, struct osmo
 		/* Nothing will be sent anyway... */
 		return;
 	}
-
-	/* Add the routing context, if it is configured */
-	if (as->cfg.routing_key.context) {
-		npar->presence |= NOTIFY_PAR_P_ROUTE_CTX;
-		npar->route_ctx = as->cfg.routing_key.context;
-	}
 }
 
-/* is the given AS one with a single ASP of IPA type? */
-static bool is_single_ipa_asp(struct osmo_ss7_as *as)
-{
-	unsigned int asp_count = 0;
-	int i;
-
-	for (i = 0; i < ARRAY_SIZE(as->cfg.asps); i++) {
-		struct osmo_ss7_asp *asp = as->cfg.asps[i];
-		if (!asp)
-			continue;
-		asp_count++;
-		if (asp->cfg.proto != OSMO_SS7_ASP_PROT_IPA)
-			return false;
-	}
-	if (asp_count == 1)
-		return true;
-	return false;
-}
-
-static void ipa_add_route(struct osmo_fsm_inst *fi)
-{
-	struct xua_as_fsm_priv *xafp = (struct xua_as_fsm_priv *) fi->priv;
-	struct osmo_ss7_as *as = xafp->as;
-	struct osmo_ss7_instance *inst = as->inst;
-
-	if (ss7_route_table_find_route_by_dpc_mask(inst->rtable_system, as->cfg.routing_key.pc, 0xffffff))
-		return;
-
-	/* As opposed to M3UA, there is no RKM and we have to implicitly
-	 * automatically add a route once an IPA connection has come up */
-	if (ss7_route_create(inst->rtable_system, as->cfg.routing_key.pc, 0xffffff, as->cfg.name))
-		xafp->ipa_route_created = true;
-}
-
-static void ipa_del_route(struct osmo_fsm_inst *fi)
-{
-	struct xua_as_fsm_priv *xafp = (struct xua_as_fsm_priv *) fi->priv;
-	struct osmo_ss7_as *as = xafp->as;
-	struct osmo_ss7_instance *inst = as->inst;
-	struct osmo_ss7_route *rt;
-
-	/* don't delete a route if we added none */
-	if (!xafp->ipa_route_created)
-		return;
-
-	/* find the route which we have created if we ever reached ipa_asp_fsm_wait_id_ack2 */
-	rt = ss7_route_table_find_route_by_dpc_mask(inst->rtable_system, as->cfg.routing_key.pc, 0xffffff);
-	/* no route found, bail out */
-	if (!rt) {
-		LOGPFSML(fi, LOGL_NOTICE, "Attempting to delete route for this IPA AS, but cannot "
-			 "find route for DPC %s. Did you manually delete it?\n",
-			 osmo_ss7_pointcode_print(inst, as->cfg.routing_key.pc));
-		return;
-	}
-
-	/* route points to different AS, bail out */
-	if (rt->dest.as != as) {
-		LOGPFSML(fi, LOGL_NOTICE, "Attempting to delete route for this IPA ASP, but found "
-			 "route for DPC %s points to different AS (%s)\n",
-			 osmo_ss7_pointcode_print(inst, as->cfg.routing_key.pc), rt->dest.as->cfg.name);
-		return;
-	}
-
-	ss7_route_destroy(rt);
-	xafp->ipa_route_created = false;
-}
-
-
-
 /* is any other ASP in this AS in state != DOWN? */
 static bool check_any_other_asp_not_down(struct osmo_ss7_as *as, struct osmo_ss7_asp *asp_cmp)
 {
@@ -370,8 +321,8 @@ static void notify_any_other_active_asp_as_inactive(struct osmo_ss7_as *as, stru
 		.status_info = M3UA_NOTIFY_I_OT_ALT_ASP_ACT,
 	};
 
-	if (asp_cmp->asp_id_present)
-		npar.asp_id = asp_cmp->asp_id;
+	if (asp_cmp->remote_asp_id_present)
+		npar.asp_id = asp_cmp->remote_asp_id;
 
 	for (i = 0; i < ARRAY_SIZE(as->cfg.asps); i++) {
 		struct osmo_ss7_asp *asp = as->cfg.asps[i];
@@ -411,31 +362,139 @@ static void xua_as_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *data
 	}
 }
 
+/* AS became inactive, trigger SNM PC unavailability for PCs it served: */
+static void as_snm_pc_unavailable(struct osmo_ss7_as *as)
+{
+	struct osmo_ss7_instance *s7i = as->inst;
+	struct osmo_ss7_route_table *rtbl = s7i->rtable_system;
+	struct osmo_ss7_combined_linkset *clset;
+	struct osmo_ss7_route *rt;
+	uint32_t aff_pc;
+
+	LOGPAS(as, DLSS7, LOGL_INFO, "AS became inactive, some routes to remote PCs may have become unavailable\n");
+	/* we assume the combined_links are sorted by mask length, i.e. more
+	 * specific combined links first, and less specific combined links with shorter
+	 * mask later */
+	llist_for_each_entry(clset, &rtbl->combined_linksets, list) {
+		llist_for_each_entry(rt, &clset->routes, list) {
+			if (rt->dest.as != as)
+				continue;
+			if (rt->status == OSMO_SS7_ROUTE_STATUS_UNAVAILABLE)
+				continue; /* Route was not available before regardless of AS state... */
+			if (ss7_route_is_summary(rt))
+				continue; /* Only announce changes for fully qualified routes */
+			/* We found a PC served by the AS which went down. If no
+			 * alternative route is still in place now towards that PC, announce unavailability
+			 * to upper layers. */
+			if (ss7_route_table_dpc_is_accessible(rtbl, rt->cfg.pc))
+				continue;
+			LOGPAS(as, DLSS7, LOGL_INFO, "AS became inactive => remote pc=%u=%s is now unavailable\n",
+			       rt->cfg.pc, osmo_ss7_pointcode_print(s7i, rt->cfg.pc));
+			aff_pc = htonl(rt->cfg.pc);
+			xua_snm_pc_available(as, &aff_pc, 1, NULL, false);
+		}
+	}
+
+	/* Generate PC unavailability indications for PCs in the sccp address-book.
+	 * These are generally PCs the user wants to get information about, and
+	 * they may have no fully-qualified route because:
+	 * - user may have not configured such fully qualified route, but a
+	 *   summary route (eg. 0.0.0/0 default route).
+	 * - peer SG may have not announced them (DAVA/DUNA).
+	 */
+	struct osmo_sccp_addr_entry *entry;
+	llist_for_each_entry(entry, &s7i->cfg.sccp_address_book, list) {
+		if (!(entry->addr.presence & OSMO_SCCP_ADDR_T_PC))
+			continue;
+		if (osmo_ss7_pc_is_local(s7i, entry->addr.pc))
+			continue;
+		/* Still available: */
+		if (ss7_route_table_dpc_is_accessible_skip_as(rtbl, entry->addr.pc, as))
+			continue;
+		LOGPAS(as, DLSS7, LOGL_INFO, "AS became inactive => address-book pc=%u=%s is now unavailable\n",
+		       entry->addr.pc, osmo_ss7_pointcode_print(s7i, entry->addr.pc));
+		aff_pc = htonl(entry->addr.pc);
+		xua_snm_pc_available(as, &aff_pc, 1, NULL, false);
+	}
+}
+
+/* AS became active, trigger SNM PC availability for PCs it served: */
+static void as_snm_pc_available(struct osmo_ss7_as *as)
+{
+	struct osmo_ss7_instance *s7i = as->inst;
+	struct osmo_ss7_route_table *rtbl = s7i->rtable_system;
+	struct osmo_ss7_combined_linkset *clset;
+	struct osmo_ss7_route *rt;
+	uint32_t aff_pc;
+
+	LOGPAS(as, DLSS7, LOGL_INFO, "AS became active, some routes to remote PCs may have become available\n");
+	/* we assume the combined_links are sorted by mask length, i.e. more
+	 * specific combined links first, and less specific combined links with shorter
+	 * mask later */
+	llist_for_each_entry(clset, &rtbl->combined_linksets, list) {
+		llist_for_each_entry(rt, &clset->routes, list) {
+			if (rt->dest.as != as)
+				continue;
+			if (ss7_route_is_summary(rt))
+				continue; /* Only announce changes for fully qualified routes */
+			/* We found a PC served by the AS which went up. If there's
+			 * a route now in place towards that PC, announce availability
+			 * to upper layers. */
+			if (!ss7_route_table_dpc_is_accessible(rtbl, rt->cfg.pc))
+				continue;
+			LOGPAS(as, DLSS7, LOGL_INFO, "AS became active => remote pc=%u=%s is now available\n",
+			       rt->cfg.pc, osmo_ss7_pointcode_print(s7i, rt->cfg.pc));
+			aff_pc = htonl(rt->cfg.pc);
+			xua_snm_pc_available(as, &aff_pc, 1, NULL, true);
+		}
+	}
+
+	/* Generate PC availability indications for PCs in the sccp address-book.
+	 * These are generally PCs the user wants to get information about, and
+	 * they may have no fully-qualified route because:
+	 * - user may have not configured such fully qualified route, but a
+	 *   summary route (eg. 0.0.0/0 default route).
+	 * - peer SG may have not announced them (DAVA/DUNA).
+	 */
+	struct osmo_sccp_addr_entry *entry;
+	llist_for_each_entry(entry, &s7i->cfg.sccp_address_book, list) {
+		if (!(entry->addr.presence & OSMO_SCCP_ADDR_T_PC))
+			continue;
+		if (osmo_ss7_pc_is_local(s7i, entry->addr.pc))
+			continue;
+		/* PC was already available: */
+		if (ss7_route_table_dpc_is_accessible_skip_as(rtbl, entry->addr.pc, as))
+			continue;
+		/* Try to find if there's an available route via the AS matching this PC. */
+		if (!ss7_route_table_dpc_is_accessible_via_as(rtbl, entry->addr.pc, as))
+			continue;
+		LOGPAS(as, DLSS7, LOGL_INFO, "AS became active => address-book pc=%u=%s is now available\n",
+			entry->addr.pc, osmo_ss7_pointcode_print(s7i, entry->addr.pc));
+		aff_pc = htonl(entry->addr.pc);
+		xua_snm_pc_available(as, &aff_pc, 1, NULL, true);
+	}
+}
+
 /* onenter call-back responsible of transmitting NTFY to all ASPs in
  * case of AS state changes */
 static void xua_as_fsm_onenter(struct osmo_fsm_inst *fi, uint32_t old_state)
 {
 	struct xua_as_fsm_priv *xafp = (struct xua_as_fsm_priv *) fi->priv;
 	struct osmo_ss7_as *as = xafp->as;
+	struct osmo_ss7_instance *s7i = as->inst;
 	struct osmo_xlm_prim_notify npar;
 
-	fill_notify_statchg_pars(fi, &npar);
-
 	switch (fi->state) {
 	case XUA_AS_S_INACTIVE:
 		/* continue below */
 		break;
 	case XUA_AS_S_ACTIVE:
-		if (is_single_ipa_asp(as))
-			ipa_add_route(fi);
 		/* continue below */
 		break;
 	case XUA_AS_S_PENDING:
 		/* continue below */
 		break;
 	case XUA_AS_S_DOWN:
-		if (is_single_ipa_asp(as))
-			ipa_del_route(fi);
 		/* RFC4666 sec 4.3.2 AS States:
 		   If we end up here, it means no ASP is ACTIVE or INACTIVE,
 		   meaning no ASP can have already configured the traffic mode
@@ -456,15 +515,31 @@ static void xua_as_fsm_onenter(struct osmo_fsm_inst *fi, uint32_t old_state)
 
 	as_notify_all_asp(xafp->as, &npar);
 
-	/* only if we are the SG, we must start broadcasting availability information
-	 * to everyone else */
-	if (get_local_role(xafp->as) == OSMO_SS7_ASP_ROLE_SG) {
+	bool became_available = (old_state != XUA_AS_S_ACTIVE && fi->state == XUA_AS_S_ACTIVE);
+	bool became_unavailable = (old_state == XUA_AS_S_ACTIVE && fi->state != XUA_AS_S_ACTIVE);
+	int role = ss7_as_get_local_role(xafp->as);
+
+	switch (role) {
+	case OSMO_SS7_ASP_ROLE_ASP:
+		if (s7i->sccp) {
+			if (became_available)
+				as_snm_pc_available(as);
+			else if (became_unavailable)
+				as_snm_pc_unavailable(as);
+		}
+		break;
+	case OSMO_SS7_ASP_ROLE_SG:
+		/* only if we are the SG, we must start broadcasting availability information
+		 * to everyone else */
 		/* advertise availability of the routing key to others */
-		uint32_t aff_pc = htonl(as->cfg.routing_key.pc);
-		if (old_state != XUA_AS_S_ACTIVE && fi->state == XUA_AS_S_ACTIVE)
-			xua_snm_pc_available(as, &aff_pc, 1, NULL, true);
-		else if (old_state == XUA_AS_S_ACTIVE && fi->state != XUA_AS_S_ACTIVE)
-			xua_snm_pc_available(as, &aff_pc, 1, NULL, false);
+		if (became_available || became_unavailable) {
+			uint32_t aff_pc = htonl(as->cfg.routing_key.pc);
+			xua_snm_pc_available(as, &aff_pc, 1, NULL, became_available);
+		}
+		break;
+	case OSMO_SS7_ASP_ROLE_IPSP:
+		/* TODO */
+		break;
 	}
 };
 
@@ -485,8 +560,11 @@ static void xua_as_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void *
 			osmo_fsm_inst_state_chg(fi, XUA_AS_S_DOWN, 0, 0);
 		break;
 	case XUA_ASPAS_ASP_ACTIVE_IND:
+		asp = data;
 		/* one ASP transitions into ASP-ACTIVE */
 		osmo_fsm_inst_state_chg(fi, XUA_AS_S_ACTIVE, 0, 0);
+		if (asp->cfg.role == OSMO_SS7_ASP_ROLE_SG)
+			as_tx_duna_during_asp_act(xafp->as, asp);
 		break;
 	case XUA_ASPAS_ASP_INACTIVE_IND:
 		inact_ind_pars = data;
@@ -508,14 +586,8 @@ static void xua_as_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *da
 
 	switch (event) {
 	case XUA_ASPAS_ASP_DOWN_IND:
-	case XUA_ASPAS_ASP_INACTIVE_IND:
-		inact_ind_pars = data;
-		if (check_any_other_asp_in_active(xafp->as, inact_ind_pars->asp)) {
-			if (event == XUA_ASPAS_ASP_INACTIVE_IND && inact_ind_pars->asp_requires_notify) {
-				fill_notify_statchg_pars(fi, &npar);
-				tx_notify(inact_ind_pars->asp, &npar);
-			} /* ASP_DOWN_IND: ignore, nothing to be sent */
-		} else {
+		asp = data;
+		if (!check_any_other_asp_in_active(xafp->as, asp)) {
 			uint32_t recovery_msec = xafp->as->cfg.recovery_timeout_msec;
 			osmo_fsm_inst_state_chg(fi, XUA_AS_S_PENDING, 0, 0);
 			/* Start T(r) */
@@ -524,11 +596,28 @@ static void xua_as_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *da
 			 * recovery or T(r) expiry */
 		}
 		break;
+	case XUA_ASPAS_ASP_INACTIVE_IND:
+		inact_ind_pars = data;
+		if (!check_any_other_asp_in_active(xafp->as, inact_ind_pars->asp)) {
+			uint32_t recovery_msec = xafp->as->cfg.recovery_timeout_msec;
+			osmo_fsm_inst_state_chg(fi, XUA_AS_S_PENDING, 0, 0);
+			/* Start T(r) */
+			osmo_timer_schedule(&xafp->recovery.t_r, MSEC_TO_S_US(recovery_msec));
+			/* FIXME: Queue all signalling messages until
+			 * recovery or T(r) expiry */
+		} else if (inact_ind_pars->asp_requires_notify) {
+			fill_notify_statchg_pars(fi, &npar);
+			tx_notify(inact_ind_pars->asp, &npar);
+		}
+		break;
 	case XUA_ASPAS_ASP_ACTIVE_IND:
 		asp = data;
 		/* RFC466 sec 4.3.4.3 ASP Active Procedures*/
 		if (xafp->as->cfg.mode == OSMO_SS7_AS_TMOD_OVERRIDE)
 			notify_any_other_active_asp_as_inactive(xafp->as, asp);
+		/* SG role: No need to send DUNA for unknown destinations here
+		 * (see as_tx_duna_during_asp_act()), since the AS was already
+		 * active so the peer should know current status. */
 		break;
 	case XUA_AS_E_TRANSFER_REQ:
 		/* message for transmission */
@@ -541,15 +630,19 @@ static void xua_as_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *da
 static void xua_as_fsm_pending(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct xua_as_fsm_priv *xafp = (struct xua_as_fsm_priv *) fi->priv;
+	struct osmo_ss7_asp *asp;
 	struct xua_as_event_asp_inactive_ind_pars *inact_ind_pars;
 	struct xua_msg *xua;
 	struct osmo_xlm_prim_notify npar;
 
 	switch (event) {
 	case XUA_ASPAS_ASP_ACTIVE_IND:
+		asp = data;
 		/* one ASP transitions into ASP-ACTIVE */
 		osmo_timer_del(&xafp->recovery.t_r);
 		osmo_fsm_inst_state_chg(fi, XUA_AS_S_ACTIVE, 0, 0);
+		if (asp->cfg.role == OSMO_SS7_ASP_ROLE_SG)
+			as_tx_duna_during_asp_act(xafp->as, asp);
 		/* push out any pending queued messages */
 		while (!llist_empty(&xafp->recovery.queued_xua_msgs)) {
 			struct xua_msg *xua;
@@ -663,6 +756,8 @@ struct osmo_fsm_inst *xua_as_fsm_start(struct osmo_ss7_as *as, int log_level)
 	struct xua_as_fsm_priv *xafp;
 
 	fi = osmo_fsm_inst_alloc(&xua_as_fsm, as, NULL, log_level, as->cfg.name);
+	if (!fi)
+		return NULL;
 
 	xafp = talloc_zero(fi, struct xua_as_fsm_priv);
 	if (!xafp) {
